__all__ = [
    "AnyConversion",
    "Conversion",
    "LazyConversion",
    "as_names",
    "as_str",
    "catch_value_error",
    "deserializer",
    "reset_deserializers",
    "reset_serializer",
    "serializer",
]
from .conversions import AnyConversion, Conversion, LazyConversion
from .converters import (
    as_names,
    as_str,
    catch_value_error,
    deserializer,
    reset_deserializers,
    reset_serializer,
    serializer,
)



import re
from dataclasses import dataclass, replace
from typing import (
    Any,
    Callable,
    Dict,
    Literal,
    Mapping,
    Optional,
    Pattern,
    Sequence,
    TypeVar,
    Union,
)

from apischema.constraints import Constraints
from apischema.metadata.keys import SCHEMA_METADATA
from apischema.types import AnyType, MetadataMixin, Number, Undefined
from apischema.typing import is_annotated
from apischema.utils import merge_opts, replace_builtins, to_camel_case

T = TypeVar("T")
Extra = Union[Mapping[str, Any], Callable[[Dict[str, Any]], None]]

Deprecated = Union[bool, str]
ContentEncoding = Literal["7bit", "8bit", "binary", "quoted-printable", "base64"]


@dataclass(frozen=True)
class Schema(MetadataMixin):
    key = SCHEMA_METADATA
    title: Optional[str] = None
    description: Optional[str] = None
    # use a callable wrapper in order to be hashable
    default: Optional[Callable[[], Any]] = None
    examples: Optional[Sequence[Any]] = None
    format: Optional[str] = None
    deprecated: Optional[Deprecated] = None
    media_type: Optional[str] = None
    encoding: Optional[ContentEncoding] = None
    constraints: Optional[Constraints] = None
    extra: Optional[Callable[[Dict[str, Any]], None]] = None
    override: bool = False
    child: Optional["Schema"] = None

    def __call__(self, tp: T) -> T:
        if is_annotated(tp):
            raise TypeError("Cannot register schema on Annotated type")
        _schemas[replace_builtins(tp)] = self
        return tp

    def merge_into(self, base_schema: Dict[str, Any]):
        if self.override:
            base_schema.clear()
        elif self.child is not None:
            self.child.merge_into(base_schema)
        if self.constraints is not None:
            self.constraints.merge_into(base_schema)
        if self.deprecated:
            base_schema["deprecated"] = bool(self.deprecated)
        for k in ("title", "description", "examples", "format"):
            if getattr(self, k) is not None:
                base_schema[k] = getattr(self, k)
        for k in ("media_type", "encoding"):
            if getattr(self, k) is not None:
                base_schema[to_camel_case("content_" + k)] = getattr(self, k)
        if self.default is not None:
            base_schema["default"] = self.default()
        if self.extra is not None:
            self.extra(base_schema)


def schema(
    *,
    # annotations
    title: Optional[str] = None,
    description: Optional[str] = None,
    default: Any = Undefined,
    examples: Optional[Sequence[Any]] = None,
    deprecated: Optional[Deprecated] = None,
    # number
    min: Optional[Number] = None,
    max: Optional[Number] = None,
    exc_min: Optional[Number] = None,
    exc_max: Optional[Number] = None,
    mult_of: Optional[Number] = None,
    # string
    format: Optional[str] = None,
    media_type: Optional[str] = None,
    encoding: Optional[ContentEncoding] = None,
    min_len: Optional[int] = None,
    max_len: Optional[int] = None,
    pattern: Optional[Union[str, Pattern]] = None,
    # array
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
    unique: Optional[bool] = None,
    # objects
    min_props: Optional[int] = None,
    max_props: Optional[int] = None,
    # extra
    extra: Optional[Extra] = None,
    override: bool = False,
) -> Schema:
    default = None if default is Undefined else (lambda d=default: d)
    if pattern is not None:
        pattern = re.compile(pattern)
    if isinstance(extra, Mapping):
        extra = lambda js, to_update=extra: js.update(to_update)  # type: ignore
    constraints = Constraints(
        min=min,
        max=max,
        exc_min=exc_min,
        exc_max=exc_max,
        mult_of=mult_of,
        min_len=min_len,
        max_len=max_len,
        pattern=pattern,
        min_items=min_items,
        max_items=max_items,
        unique=unique,
        min_props=min_props,
        max_props=max_props,
    )
    return Schema(
        title=title,
        description=description,
        default=default,
        examples=examples,
        format=format,
        deprecated=deprecated,
        media_type=media_type,
        encoding=encoding,
        constraints=constraints,
        extra=extra,
        override=override,
    )


_schemas: Dict[Any, Schema] = {}


def get_schema(tp: AnyType) -> Optional[Schema]:
    tp = replace_builtins(tp)
    try:
        return _schemas.get(tp)
    except TypeError:
        return None


@merge_opts
def merge_schema(default: Schema, override: Schema) -> Schema:
    if override.override:
        return override
    return replace(override, child=merge_schema(default, override.child))



"""Kind of typing_extensions for this package"""
__all__ = ["get_args", "get_origin", "get_type_hints"]

import sys
from types import ModuleType, new_class
from typing import Any, Callable, Dict, Generic, Protocol, TypeVar, Union


class _FakeType:
    pass


if sys.version_info >= (3, 9):  # pragma: no cover
    from typing import Annotated, get_args, get_origin, get_type_hints
else:  # pragma: no cover
    try:
        from typing_extensions import Annotated
    except ImportError:
        pass
    try:
        from typing_extensions import get_type_hints as gth
    except ImportError:
        from typing import get_type_hints as _gth

        def gth(obj, globalns=None, localns=None, include_extras=False):
            return _gth(obj, globalns, localns)

    def get_type_hints(obj, globalns=None, localns=None, include_extras=False):
        # TODO This has been fixed in recent 3.7 and 3.8
        # fix https://bugs.python.org/issue37838
        if not isinstance(obj, (type, ModuleType)) and globalns is None:
            nsobj = obj
            while hasattr(nsobj, "__wrapped__"):
                nsobj = nsobj.__wrapped__
            globalns = getattr(nsobj, "__globals__", None)
        localns = {"unicode": str, **(localns or {})}
        return gth(obj, globalns, localns, include_extras)

    try:
        from typing_extensions import get_args, get_origin
    except ImportError:

        def get_origin(tp):
            if isinstance(tp, _AnnotatedAlias):
                return None if tp.__args__ is None else Annotated
            if tp is Generic:
                return Generic
            return getattr(tp, "__origin__", None)

        def get_args(tp):
            if isinstance(tp, _AnnotatedAlias):
                return () if tp.__args__ is None else (tp.__args__[0], *tp.__metadata__)
            res = getattr(tp, "__args__", ())
            if get_origin(tp) is Callable and res[0] is not Ellipsis:
                res = (list(res[:-1]), res[-1])
            return res


if sys.version_info >= (3, 13):
    from typing import _collect_type_parameters
elif sys.version_info >= (3, 11):
    from typing import _collect_parameters as _collect_type_parameters  # type: ignore
else:
    from typing import _collect_type_vars as _collect_type_parameters


def _generic_mro(result, tp):
    origin = get_origin(tp)
    if origin is None:
        origin = tp
    result[origin] = tp
    if hasattr(origin, "__orig_bases__"):
        parameters = _collect_type_parameters(origin.__orig_bases__)
        substitution = dict(zip(parameters, get_args(tp)))
        for base in origin.__orig_bases__:
            if get_origin(base) in result:
                continue
            base_parameters = getattr(base, "__parameters__", ())
            if base_parameters:
                base = base[tuple(substitution.get(p, p) for p in base_parameters)]
            _generic_mro(result, base)


# sentinel value to avoid to subscript Generic and Protocol
BASE_GENERIC_MRO = {Generic: Generic, Protocol: Protocol}


def generic_mro(tp):
    origin = get_origin(tp)
    if origin is None and not hasattr(tp, "__orig_bases__"):
        if not isinstance(tp, type):
            raise TypeError(f"{tp!r} is not a type or a generic alias")
        return tp.__mro__
    result = BASE_GENERIC_MRO.copy()
    _generic_mro(result, tp)
    cls = origin if origin is not None else tp
    return tuple(result.get(sub_cls, sub_cls) for sub_cls in cls.__mro__)


def resolve_type_hints(obj: Any) -> Dict[str, Any]:
    """Wrap get_type_hints to resolve type vars in case of generic inheritance.

    `obj` can also be a parametrized generic class."""
    origin_or_obj = get_origin(obj) or obj
    if isinstance(origin_or_obj, type):
        hints = {}
        for base in reversed(generic_mro(obj)):
            base_origin = get_origin(base) or base
            base_annotations = getattr(base_origin, "__dict__", {}).get(
                "__annotations__", {}
            )
            substitution = dict(
                zip(getattr(base_origin, "__parameters__", ()), get_args(base))
            )
            for name, hint in get_type_hints(base_origin, include_extras=True).items():
                if name not in base_annotations:
                    continue
                if isinstance(hint, TypeVar):
                    hints[name] = substitution.get(hint, hint)
                elif getattr(hint, "__parameters__", ()):
                    hints[name] = (Union if is_union(hint) else hint)[
                        tuple(substitution.get(p, p) for p in hint.__parameters__)
                    ]
                else:
                    hints[name] = hint
        return hints
    else:
        return get_type_hints(obj, include_extras=True)


_T = TypeVar("_T")
_GenericAlias: Any = type(Generic[_T])
try:
    _AnnotatedAlias: Any = type(Annotated[_T, ...])
except NameError:
    _AnnotatedAlias = _FakeType


def is_new_type(tp: Any) -> bool:
    return hasattr(tp, "__supertype__")


def is_annotated(tp: Any) -> bool:
    try:
        from typing import Annotated

        return get_origin(tp) == Annotated
    except ImportError:
        try:
            from typing_extensions import Annotated  # type: ignore

            return get_origin(tp) == Annotated
        except ImportError:
            return False


def is_literal(tp: Any) -> bool:
    from typing import Literal

    origin = get_origin(tp)
    if origin is Literal:
        return True
    try:
        from typing_extensions import Literal as Literal2

        return get_origin(tp) is Literal2
    except ImportError:
        return False


def is_literal_string(tp: Any) -> bool:
    try:
        from typing import LiteralString

        if tp is LiteralString:
            return True
    except ImportError:
        pass
    try:
        from typing_extensions import LiteralString

        return tp is LiteralString
    except ImportError:
        return False


def is_named_tuple(tp: Any) -> bool:
    return issubclass(tp, tuple) and hasattr(tp, "_fields")


def is_typed_dict(tp: Any) -> bool:
    # TODO use python 3.10 typing.is_typeddict
    from typing import TypedDict

    if isinstance(tp, type(new_class("_TypedDictImplem", (TypedDict,)))):
        return True
    try:
        from typing_extensions import TypedDict as TypedDict2

        return isinstance(tp, type(new_class("_TypedDictImplem", (TypedDict2,))))
    except ImportError:
        return False


def is_type_var(tp: Any) -> bool:
    return isinstance(tp, TypeVar)


# py38 get_origin of builtin wrapped generics return the unsubscriptable builtin
# type.
if sys.version_info < (3, 9):
    import typing

    TYPING_ALIASES = {
        getattr(elt, "__origin__", None): elt for elt in typing.__dict__.values()
    }

    def typing_origin(origin: Any) -> Any:
        return TYPING_ALIASES.get(origin, origin)

else:
    typing_origin = lambda tp: tp


def is_type(tp: Any) -> bool:
    """isinstance is not enough because in py39: isinstance(list[int], type) == True"""
    return isinstance(tp, type) and not get_args(tp)


def is_union(tp: Any) -> bool:
    try:
        from types import UnionType

        return tp in (UnionType, Union)
    except ImportError:
        return tp is Union



import os
import platform

from setuptools import Extension, setup

ext_modules = [
    Extension(
        f"apischema.{package}.methods",
        sources=[f"apischema/{package}/methods.c"],
        optional=True,
    )
    for package in ("deserialization", "serialization")
    # Cythonization makes apischema slower using PyPy
    if platform.python_implementation() != "PyPy" and "NO_EXTENSION" not in os.environ
]
setup(ext_modules=ext_modules)



import operator
import re
import sys
from base64 import b64decode, b64encode
from collections import deque
from datetime import date, datetime, time
from decimal import Decimal
from ipaddress import (
    IPv4Address,
    IPv4Interface,
    IPv4Network,
    IPv6Address,
    IPv6Interface,
    IPv6Network,
)
from pathlib import (
    Path,
    PosixPath,
    PurePath,
    PurePosixPath,
    PureWindowsPath,
    WindowsPath,
)
from typing import TypeVar
from uuid import UUID

from apischema import ValidationError, deserializer, schema, serializer, type_name
from apischema.conversions import Conversion, as_str, catch_value_error

# =================== bytes =====================

deserializer(Conversion(b64decode, source=str, target=bytes))


@serializer
def to_base64(b: bytes) -> str:
    return b64encode(b).decode()


type_name(graphql="Bytes")(bytes)
schema(encoding="base64")(bytes)


# ================ collections ==================

T = TypeVar("T")
if sys.version_info >= (3, 10):
    deserializer(Conversion(deque, source=list[T], target=deque[T]))  # type: ignore
    serializer(Conversion(list, source=deque[T], target=list[T]))  # type: ignore
else:
    from typing import Deque, List

    deserializer(Conversion(deque, source=List[T], target=Deque[T]))
    serializer(Conversion(list, source=Deque[T], target=List[T]))


# ================== datetime ===================

for cls, format in [(date, "date"), (datetime, "date-time"), (time, "time")]:
    fromisoformat = catch_value_error(cls.fromisoformat)  # type: ignore
    deserializer(Conversion(fromisoformat, source=str, target=cls))
    serializer(Conversion(cls.isoformat, source=cls, target=str))  # type: ignore
    type_name(graphql=cls.__name__.capitalize())(cls)
    schema(format=format)(cls)

# ================== decimal ====================

deserializer(Conversion(catch_value_error(Decimal), source=float, target=Decimal))
serializer(Conversion(float, source=Decimal, target=float))
type_name(None)(Decimal)

# ================= ipaddress ===================

for classes, format in [
    ((IPv4Address, IPv4Interface, IPv4Network), "ipv4"),
    ((IPv6Address, IPv6Interface, IPv6Network), "ipv6"),
]:
    for cls in classes:
        as_str(cls)
        type_name(graphql=cls.__name__)(cls)
        schema(format=format)(cls)

# ==================== path =====================

for cls in (PurePath, PurePosixPath, PureWindowsPath, Path, PosixPath, WindowsPath):
    as_str(cls)
    type_name(None)(cls)

# =================== pattern ===================


@deserializer
def _compile(pattern: str) -> re.Pattern:
    try:
        return re.compile(pattern)
    except re.error as err:
        raise ValidationError(str(err))


serializer(Conversion(operator.attrgetter("pattern"), source=re.Pattern, target=str))
type_name(None)(re.Pattern)

# ==================== uuid =====================

as_str(UUID)
type_name(graphql="UUID")
schema(format="uuid")(UUID)
